#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <FLAC/file_decoder.h>

int dsp_fd;
int channels;
short *outbuf;
bool started_playing;
char *filename;

FLAC__StreamDecoderWriteStatus write_cb( const FLAC__FileDecoder *decoder,
	const FLAC__Frame *frame, const FLAC__int32 * const buffer[],
	void *client_data ) {
  int i, j, pos;

  if ( !started_playing ) {
	printf( "Playing stream from %s...\n", filename );
	started_playing = true;
  }

  pos = 0;
  for ( i = 0; i < frame->header.blocksize; i++ )
  	for ( j = 0; j < channels; j++ )
		outbuf[pos++] = buffer[j][i];

  write( dsp_fd, outbuf, pos * sizeof(short) );
  
  return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

bool parse_vorbis_comment( const FLAC__StreamMetadata_VorbisComment_Entry *entry,
	char *tagname, int taglen, char *value, int vallen  ) {
  int count;
  char *data, *eqpos, *zpos;

  data = (char*)entry->entry;
  count = entry->length;
  if ( count > taglen ) count = taglen;

  eqpos = (char*) memchr( data, '=', count );
  if ( !eqpos ) return false;

  memcpy( tagname, data, eqpos - data );
  tagname[eqpos - data] = 0;

  count = entry->length - (eqpos - data + 1);
  if ( count > vallen ) count = vallen;
  
  data = eqpos + 1;
  zpos = (char*) memchr( data, 0, count );
  if ( !zpos ) zpos = data + count;

  memcpy( value, data, zpos - data );
  value[zpos - data] = 0;

  return true;
}

void meta_cb( const FLAC__FileDecoder *decoder,
	const FLAC__StreamMetadata *metadata, void *client_data ) {
  int i, val;
  char tag[20], tval[40];
  bool rc;

  if ( metadata->type == FLAC__METADATA_TYPE_VORBIS_COMMENT ) {
	const FLAC__StreamMetadata_VorbisComment *vc;
	char tag[20], tval[40];
	char *artist = 0, *album = 0, *song = 0, *year = 0;
	bool rc;

	vc = &metadata->data.vorbis_comment;
	if ( !vc ) return;

	for ( i = 0; i < vc->num_comments; i++ ) {
		rc = parse_vorbis_comment( &vc->comments[i],
			 tag, sizeof(tag), tval, sizeof(tval) );
		if ( !rc ) continue;
		
		if ( !strcmp( tag, "artist" )  ) artist = strdup( tval );
		else if ( !strcmp( tag, "album" )  ) album = strdup( tval );
		else if ( !strcmp( tag, "song" )  ) song = strdup( tval );
		else if ( !strcmp( tag, "year" )  ) year = strdup( tval );
	}

	if ( artist || album || year || song ) {
		printf( "Artist: %s\n", artist? artist: "" );
		printf( "Album : %s\n", album? album: "" );
		printf( "Year  : %s\n", year? year: "" );
		printf( "Song  : %s\n", song? song: "" );
		putchar( '\n' );

		free( artist );
		free( album );
		free( year );
		free( song );
	}

  	return;
  }

  if ( metadata->type != FLAC__METADATA_TYPE_STREAMINFO ) return;

  val = metadata->data.stream_info.bits_per_sample;
/*if ( val == 8 )
  	val = AFMT_S8;
  else */ if ( val == 16 )
	val = AFMT_S16_LE;
  else {
  	printf( "%d bits per sample format not supported\n", val );
	exit( 1 );
  }

  ioctl( dsp_fd, SNDCTL_DSP_SETFMT, &val );

  val = metadata->data.stream_info.channels;
  channels = val;
  if ( val == 1 )
  	val = 0;
  else if ( val == 2 )
  	val = 1;
  else {
  	printf( "%d channels format not supported\n", val );
	exit( 1 );
  }

  ioctl( dsp_fd, SNDCTL_DSP_STEREO, &val );

  val = metadata->data.stream_info.sample_rate;
  ioctl( dsp_fd, SNDCTL_DSP_SPEED, &val );
}

void error_cb( const FLAC__FileDecoder *decoder,
	FLAC__StreamDecoderErrorStatus status, void *client_data ) {
  puts( "error cb" );
}

int main( int argc, char **argv ) {
  FLAC__FileDecoder *fd;
  int rc, idx;

  if ( argc < 2 ) {
  	puts( "Usage: flacplay <filename.flac> ..." );
	return 1;
  }

  dsp_fd = open( "/dev/dsp", O_WRONLY );
  if ( dsp_fd == -1 ) {
	perror( "open /dev/dsp" );
	return 1;
  }
// two channels at most
  outbuf = (short*) malloc( FLAC__MAX_BLOCK_SIZE * 2 * sizeof(short) );

  fd = FLAC__file_decoder_new();

  for ( idx = 1; idx < argc; idx++ ) {
	started_playing = false;
	filename = argv[idx];
	FLAC__file_decoder_set_filename( fd, filename );
	FLAC__file_decoder_set_metadata_respond( fd,
			FLAC__METADATA_TYPE_STREAMINFO );
	FLAC__file_decoder_set_metadata_respond( fd,
			FLAC__METADATA_TYPE_VORBIS_COMMENT );
	FLAC__file_decoder_set_write_callback( fd, write_cb );
	FLAC__file_decoder_set_metadata_callback( fd, meta_cb );
	FLAC__file_decoder_set_error_callback( fd, error_cb );

	rc = FLAC__file_decoder_init( fd );
	if ( rc != FLAC__FILE_DECODER_OK ) {
		puts( "failed to initialize decoder" );
		return 1;
	}

	rc = FLAC__file_decoder_process_until_end_of_file( fd );
	if ( rc == false ) {
		puts( "decoding failure" );
		return 1;
	}

	FLAC__file_decoder_finish( fd );
  }

  FLAC__file_decoder_delete( fd );

  free( outbuf );
  close( dsp_fd );

  return 0;
}

